package de.thdeg.greiner.superpangworld.graphics.base;

import de.thdeg.greiner.superpangworld.gameview.GameView;
import de.thdeg.greiner.superpangworld.graphics.helper.HelperValues;
import de.thdeg.greiner.superpangworld.graphics.moveable.Harpoon;
import de.thdeg.greiner.superpangworld.graphics.moveable.PlayerObject;

import java.awt.geom.Point2D;
import java.util.*;

/**
 * A Bubble, which moves around on the screen and can be shot by the player.
 */
public abstract class Bubble extends CollidingGameObject implements MovingGameObject{

    private static final Point2D VELOCITY_RIGHT = new Point2D.Double(3,1);
    private static final Point2D VELOCITY_LEFT = new Point2D.Double(-3,1);

    /** Random generator */
    private final Random random;
    /** Flag, if the bubble is in the spawning phase */
    private boolean spawning;
    /** Spawning speed */
    private int spawnSpeed;
    /** Unique id for the spawn event timer */
    private final long spawnID;
    /** Temporary movement action */
    private Position nextPosition;

    private Point2D velocity;

    /**
     * Create a bubble with default values.
     *
     * @param gameView the {@link GameView} to display the bubble
     * @param objectsToCollideWith the list of {@link CollidableGameObject} the bubble can collide with
     */
    public Bubble(GameView gameView, ArrayList<CollidableGameObject> objectsToCollideWith){
        super(gameView,objectsToCollideWith);
        rotation = 90;
        size = 10;
        width = (int) (12 * size);
        height = (int) (12 * size);
        speedInPixel = 1;

        spawnID = UUID.randomUUID().getLeastSignificantBits();
        random = new Random();
        position.setTo(random.nextInt(GameView.WIDTH-width),-width);
        spawning = true;
        spawnSpeed = 500;
        gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);

        velocity = new Point2D.Double(-3.5,1.4);

    }

    /**
     * Create a bubble.
     *
     * @param gameView the {@link GameView} to display the bubble
     * @param objectsToCollideWith the list of {@link CollidableGameObject} the bubble can collide with
     * @param size the object size
     * @param speedInPixel the speed in pixel per tick
     * @param position the position
     * @param spawning flag, if the bubble is in the spawning phase
     * @param moveLeftToRight flag, if the bubble is moving left to right
     */
    public Bubble(GameView gameView, ArrayList<CollidableGameObject> objectsToCollideWith, double size, double speedInPixel, Position position, boolean spawning, boolean moveLeftToRight){
        super(gameView,objectsToCollideWith);
        rotation = 90;
        this.size = size;
        width = (int) (12 * size);
        height = (int) (12 * size);
        this.speedInPixel = speedInPixel;
        this.position.setTo(position.x,position.y);
        spawnID = UUID.randomUUID().getLeastSignificantBits();
        random = new Random();
        this.spawning = spawning;
        if(spawning){
            spawnSpeed = 500;
            gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);
        }

        velocity = new Point2D.Double(3.5,1.4);
    }

    /**
     * Moves the bubble a tick further.
     */
    @Override
    public void updatePosition(){
        if(spawning){
            if(position.y >= HelperValues.FRAME_BORDER_WIDTH){
                spawning = false;
            } else if(gameView.timerExpired("spawn"+spawnID,"bubble"+spawnID)){
                position.down((width + HelperValues.FRAME_BORDER_WIDTH)/4.0);
                gameView.setTimer("spawn"+spawnID,"bubble"+spawnID,spawnSpeed);
            }
        }else{

            Point2D newLocation = new Point2D.Double(position.x,position.y);
            double speedFactor = 0.2;

            newLocation = new Point2D.Double(newLocation.getX() + (velocity.getX() * speedFactor), newLocation.getY() + (velocity.getY() * speedFactor));
            velocity = new Point2D.Double(velocity.getX() + HelperValues.BUBBLE_GRAVITY.getX(), velocity.getY() + HelperValues.BUBBLE_GRAVITY.getY());

            // Wall bounce
            if((newLocation.getX() >= GameView.WIDTH - width - HelperValues.FRAME_BORDER_WIDTH) || newLocation.getX() <= HelperValues.FRAME_BORDER_WIDTH){
                velocity = new Point2D.Double(velocity.getX() * -1.0, velocity.getY());
            }

            // Bottom bounce
            if(newLocation.getY() >= HelperValues.FRAME_WINDOW_HEIGHT + HelperValues.FRAME_BORDER_WIDTH - width){
                velocity = new Point2D.Double(velocity.getX(), velocity.getY() * -0.95);
                newLocation = new Point2D.Double(newLocation.getX(), HelperValues.FRAME_WINDOW_HEIGHT + HelperValues.FRAME_BORDER_WIDTH - width);
            }

            position.setTo(newLocation.getX(), newLocation.getY());

        }
    }

    /**
     * Get the score to add when bursting the bubble.
     *
     * @return the score to add
     */
    public int getBurstScore(){
        return (int)(size*100);
    }

    /**
     * Get the size of the bubble.
     *
     * @return the size
     */
    public double getSize(){
        return this.size;
    }

    /**
     * Get the speed in pixel per tick.
     *
     * @return the speed in pixel per tick
     */
    public double getSpeedInPixel(){
        return speedInPixel;
    }

    /**
     * Get the flag if the bubble is in its spawning phase.
     *
     * @return <code>true</code> if in spawning phase, else <code>false</code>
     */
    public boolean isSpawning(){
        return spawning;
    }

    @Override
    public void reactToCollision(CollidableGameObject otherObject) {
        if(otherObject instanceof PlayerObject){
            gamePlayManager.looseLife();
        }else if(otherObject instanceof Harpoon && !spawning){
            gamePlayManager.destroy((Harpoon) otherObject);
            gamePlayManager.destroy(this);
        }
    }
}
